@@ -3,8 +3,8 @@
 /*
     This file is part of Corrade.
 
-    Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016
-              Vladimír Vondruš <mosra@centrum.cz>
+    Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+                2017, 2018 Vladimír Vondruš <mosra@centrum.cz>
 
     Permission is hereby granted, free of charge, to any person obtaining a
     copy of this software and associated documentation files (the "Software"),
@@ -44,68 +44,81 @@ namespace Corrade { namespace Containers {
 
 /**
 @brief Static array wrapper
+@tparam size_   Array size
 @tparam T       Element type
-@tparam size    Array size
 
 Provides statically-sized array wrapper with API similar to @ref Array. Useful
-as more featureful alternative to plain C arrays or `std::array`.
+as more featureful alternative to plain C arrays or @ref std::array.
 
 Usage example:
-@code
+
+@code{.cpp}
 // Create default-initialized array with 5 integers and set them to some value
-Containers::Array<int> a{5};
+Containers::StaticArray<5, int> a;
 int b = 0;
 for(auto& i: a) i = b++; // a = {0, 1, 2, 3, 4}
 
 // Create array from given values
-auto b = Containers::Array<int>::from(3, 18, -157, 0);
+Containers::StaticArray<4, int> b{3, 18, -157, 0};
 b[3] = 25; // b = {3, 18, -157, 25}
 @endcode
 
-## Array initialization
+@section Containers-StaticArray-initialization Array initialization
 
 The array is by default *default-initialized*, which means that trivial types
 are not initialized at all and default constructor is called on other types. It
 is possible to initialize the array in a different way using so-called *tags*:
 
--   @ref StaticArray(DefaultInitT) is equivalent to the default case (useful
-    when you want to make the choice appear explicit).
+-   @ref StaticArray(DefaultInitT) is equivalent to the implicit parameterless
+    constructor (useful when you want to make the choice appear explicit).
+-   @ref StaticArray(InPlaceInitT, Args&&... args) is equivalent to the
+    implicit parameteric constructor (again useful when you want to make the
+    choice appear explicit).
 -   @ref StaticArray(ValueInitT) zero-initializes trivial types and calls
     default constructor elsewhere.
--   @ref StaticArray(DirectInitT, Args...) constructs all elements of the array
-    using provided arguments.
+-   @ref StaticArray(DirectInitT, Args&&... args) constructs every element of
+    the array using provided arguments.
 -   @ref StaticArray(NoInitT) does not initialize anything and you need to call
     the constructor on all elements manually using placement new,
-    `std::uninitialized_copy` or similar. This is the dangerous option.
+    @ref std::uninitialized_copy() or similar. This is the dangerous option.
 
 Example:
-@code
-// These are equivalent
-Containers::StaticArray<5, int> a1{5};
+
+@code{.cpp}
+// These two are equivalent
+Containers::StaticArray<5, int> a1;
 Containers::StaticArray<5, int> a2{Containers::DefaultInit};
 
 // Array of 100 zeros
 Containers::StaticArray<100, int> b{Containers::ValueInit};
 
+// Array of 4 values initialized in-place (these two are equivalent)
+Containers::StaticArray<4, int> c1{3, 18, -157, 0};
+Containers::StaticArray<4, int> c2{Containers::InPlaceInit, 3, 18, -157, 0};
+
 // Array of type with no default constructor
 struct Vec3 {
     Vec3(float, float, float);
 };
-Containers::StaticArray<5, Vec3> c{Containers::DirectInit, 5.2f, 0.4f, 1.0f};
+Containers::StaticArray<5, Vec3> d{Containers::DirectInit, 5.2f, 0.4f, 1.0f};
 
 // Manual construction of each element
 struct Foo {
     Foo(int index);
 };
-Containers::StaticArray<5, Foo> d{Containers::NoInit};
+Containers::StaticArray<5, Foo> e{Containers::NoInit};
 int index = 0;
-for(Foo& f: d) new(&f) Foo(index++);
+for(Foo& f: e) new(&f) Foo(index++);
 @endcode
+
+@see @ref arrayCast(StaticArray<size, T>&)
 */
-template<std::size_t size, class T> class StaticArray {
+/* Underscore at the end to avoid conflict with member size(). It's ugly, but
+   having count instead of size_ would make the naming horribly inconsistent. */
+template<std::size_t size_, class T> class StaticArray {
     public:
         enum: std::size_t {
-            Size = size     /**< Array size */
+            Size = size_    /**< Array size */
         };
         typedef T Type;     /**< @brief Element type */
 
@@ -141,27 +154,29 @@ template<std::size_t size, class T> class StaticArray {
          * @attention Internally the destruction is done using custom deleter
          *      that explicitly calls destructor on *all elements* regardless
          *      of whether they were properly constructed or not.
-         * @see @ref NoInit, @ref StaticArray(DirectInitT, Args...)
+         * @see @ref NoInit, @ref StaticArray(DirectInitT, Args&&... args)
          */
         explicit StaticArray(NoInitT) {}
 
         /**
          * @brief Construct direct-initialized array
          *
-         * Constructs the array using the @ref StaticArray(NoInitT) "StaticArray(NoInitT)"
-         * constructor and then initializes each element with placement new
-         * using forwarded @p arguments.
-         * @see @ref StaticArray(InPlaceInitT, Args&&...)
+         * Constructs the array using the @ref StaticArray(NoInitT) constructor
+         * and then initializes each element with placement new using forwarded
+         * @p args.
+         * @see @ref StaticArray(InPlaceInitT, Args&&... args)
          */
         template<class ...Args> explicit StaticArray(DirectInitT, Args&&... args);
 
         /**
-         * @brief Construct array in-place
+         * @brief Construct in-place-initialized array
          *
          * The arguments are forwarded to the array constructor.
-         * @see @ref StaticArray(DirectInitT, Args&&...)
+         * @see @ref StaticArray(DirectInitT, Args&&... args)
          */
-        template<class ...Args> explicit StaticArray(InPlaceInitT, Args&&... args): _data{std::forward<Args>(args)...} {}
+        template<class ...Args> explicit StaticArray(InPlaceInitT, Args&&... args): _data{std::forward<Args>(args)...} {
+            static_assert(sizeof...(args) == size_, "Containers::StaticArray: wrong number of initializers");
+        }
 
         /**
          * @brief Construct default-initialized array
@@ -171,78 +186,106 @@ template<std::size_t size, class T> class StaticArray {
          */
         explicit StaticArray(): StaticArray{DefaultInit} {}
 
+        /**
+         * @brief Construct in-place-initialized array
+         *
+         * Alias to @ref StaticArray(InPlaceInitT, Args&&... args).
+         * @see @ref StaticArray(DirectInitT, Args&&... args)
+         */
+        #ifdef DOXYGEN_GENERATING_OUTPUT
+        template<class ...Args> explicit StaticArray(Args&&... args);
+        #else
+        template<class First, class ...Next> explicit StaticArray(First&& first, Next&&... next): StaticArray{InPlaceInit, std::forward<First>(first), std::forward<Next>(next)...} {}
+        #endif
+
         /** @brief Copying is not allowed */
-        StaticArray(const StaticArray<size, T>&) = delete;
+        StaticArray(const StaticArray<size_, T>&) = delete;
 
         /** @brief Moving is not allowed */
-        StaticArray(StaticArray<size, T>&&) = delete;
+        StaticArray(StaticArray<size_, T>&&) = delete;
 
         ~StaticArray();
 
         /** @brief Copying is not allowed */
-        StaticArray<size, T>& operator=(const StaticArray<size, T>&) = delete;
+        StaticArray<size_, T>& operator=(const StaticArray<size_, T>&) = delete;
 
         /** @brief Moving is not allowed */
-        StaticArray<size, T>& operator=(StaticArray<size, T>&&) = delete;
+        StaticArray<size_, T>& operator=(StaticArray<size_, T>&&) = delete;
 
         /**
          * @brief Convert to @ref ArrayView
          *
-         * Enabled only if `T*` is implicitly convertible to `U*`. Note
-         * that, similarly as with raw pointers, you need to ensure that both
-         * types have the same size.
+         * Enabled only if @cpp T* @ce is implicitly convertible to @cpp U* @ce.
+         * Expects that both types have the same size.
+         * @see @ref arrayView(StaticArray<size, T>&)
          */
         #ifdef DOXYGEN_GENERATING_OUTPUT
         template<class U>
         #else
-        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value>::type>
+        template<class U, class V = typename std::enable_if<!std::is_void<U>::value && std::is_convertible<T*, U*>::value>::type>
         #endif
-        /*implicit*/ operator ArrayView<U>() noexcept { return {_data, size}; }
-
-        /**
-         * @brief Convert to const @ref ArrayView
-         *
-         * Enabled only if `const T*` is implicitly convertible to `U*`. Note
-         * that, similarly as with raw pointers, you need to ensure that both
-         * types have the same size.
-         */
-        #ifdef DOXYGEN_GENERATING_OUTPUT
-        template<class U>
-        #else
-        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value>::type>
-        #endif
-        /*implicit*/ operator ArrayView<const U>() const noexcept { return {_data, size}; }
-
-        /**
-         * @brief Convert to @ref StaticArrayView
-         *
-         * Enabled only if `T*` is implicitly convertible to `U*`. Note
-         * that, similarly as with raw pointers, you need to ensure that both
-         * types have the same size.
-         */
-        #ifdef DOXYGEN_GENERATING_OUTPUT
-        template<class U>
-        #else
-        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value>::type>
-        #endif
-        /*implicit*/ operator StaticArrayView<size, U>() noexcept {
-            return StaticArrayView<size, U>{_data};
+        /*implicit*/ operator ArrayView<U>() noexcept {
+            static_assert(sizeof(T) == sizeof(U), "type sizes are not compatible");
+            return {_data, size_};
         }
 
         /**
          * @brief Convert to const @ref ArrayView
          *
-         * Enabled only if `const T*` is implicitly convertible to `U*`. Note
-         * that, similarly as with raw pointers, you need to ensure that both
-         * types have the same size.
+         * Enabled only if @cpp T* @ce or @cpp const T* @ce is implicitly
+         * convertible to @cpp U* @ce. Expects that both types have the same
+         * size.
+         * @see @ref arrayView(const StaticArray<size, T>&)
+         */
+        #ifdef DOXYGEN_GENERATING_OUTPUT
+        template<class U>
+        #else
+        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value || std::is_convertible<T*, const U*>::value>::type>
+        #endif
+        /*implicit*/ operator ArrayView<const U>() const noexcept {
+            static_assert(sizeof(T) == sizeof(U), "type sizes are not compatible");
+            return {_data, size_};
+        }
+
+        /** @overload */
+        /*implicit*/ operator ArrayView<const void>() const noexcept {
+            /* Yes, the size is properly multiplied by sizeof(T) by the constructor */
+            return {_data, size_};
+        }
+
+        /**
+         * @brief Convert to @ref StaticArrayView
+         *
+         * Enabled only if @cpp T* @ce is implicitly convertible to @cpp U* @ce.
+         * Expects that both types have the same size.
+         * @see @ref staticArrayView(StaticArray<size, T>&)
          */
         #ifdef DOXYGEN_GENERATING_OUTPUT
         template<class U>
         #else
         template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value>::type>
         #endif
-        /*implicit*/ operator StaticArrayView<size, const U>() const noexcept {
-            return StaticArrayView<size, const U>{_data};
+        /*implicit*/ operator StaticArrayView<size_, U>() noexcept {
+            static_assert(sizeof(T) == sizeof(U), "type sizes are not compatible");
+            return StaticArrayView<size_, U>{_data};
+        }
+
+        /**
+         * @brief Convert to const @ref StaticArrayView
+         *
+         * Enabled only if @cpp T* @ce or @cpp const T* @ce is implicitly
+         * convertible to @cpp U* @ce. Expects that both types have the same
+         * size.
+         * @see @ref staticArrayView(const StaticArray<size, T>&)
+         */
+        #ifdef DOXYGEN_GENERATING_OUTPUT
+        template<class U>
+        #else
+        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value || std::is_convertible<T*, const U*>::value>::type>
+        #endif
+        /*implicit*/ operator StaticArrayView<size_, const U>() const noexcept {
+            static_assert(sizeof(T) == sizeof(U), "type sizes are not compatible");
+            return StaticArrayView<size_, const U>{_data};
         }
 
         /* `char* a = Containers::Array<char>(5); a[3] = 5;` would result in
@@ -267,15 +310,29 @@ template<std::size_t size, class T> class StaticArray {
         T* data() { return _data; }
         const T* data() const { return _data; }             /**< @overload */
 
+        /**
+         * @brief Array size
+         *
+         * Equivalent to @ref Size.
+         */
+        constexpr std::size_t size() const { return size_; }
+
+        /**
+         * @brief Whether the array is empty
+         *
+         * Always true (it's not possible to create zero-sized C array).
+         */
+        constexpr bool empty() const { return !size_; }
+
         /** @brief Pointer to first element */
         T* begin() { return _data; }
         const T* begin() const { return _data; }            /**< @overload */
         const T* cbegin() const { return _data; }           /**< @overload */
 
         /** @brief Pointer to (one item after) last element */
-        T* end() { return _data + size; }
-        const T* end() const { return _data + size; }       /**< @overload */
-        const T* cend() const { return _data + size; }      /**< @overload */
+        T* end() { return _data + size_; }
+        const T* end() const { return _data + size_; }      /**< @overload */
+        const T* cend() const { return _data + size_; }     /**< @overload */
 
         /**
          * @brief Reference to array slice
@@ -301,7 +358,8 @@ template<std::size_t size, class T> class StaticArray {
         /**
          * @brief Fixed-size array slice
          *
-         * Both @p begin and `begin + viewSize` are expected to be in range.
+         * Both @cpp begin @ce and @cpp begin + viewSize @ce are expected to be
+         * in range.
          */
         template<std::size_t viewSize> StaticArrayView<viewSize, T> slice(T* begin) {
             return ArrayView<T>(*this).template slice<viewSize>(begin);
@@ -363,18 +421,113 @@ template<std::size_t size, class T> class StaticArray {
         #endif
 
         union {
-            T _data[size];
+            T _data[size_];
         };
 };
 
-template<std::size_t size, class T> template<class ...Args> StaticArray<size, T>::StaticArray(DirectInitT, Args&&... args): StaticArray{NoInit} {
+/** @relatesalso StaticArray
+@brief Make view on @ref StaticArray
+
+Convenience alternative to calling @ref StaticArray::operator ArrayView<U>()
+explicitly. The following two lines are equivalent:
+
+@code{.cpp}
+Containers::StaticArray<5, std::uint32_t> data;
+
+Containers::ArrayView<std::uint32_t> a{data};
+auto b = Containers::arrayView(data);
+@endcode
+*/
+template<std::size_t size, class T> constexpr ArrayView<T> arrayView(StaticArray<size, T>& array) {
+    return ArrayView<T>{array};
+}
+
+/** @relatesalso StaticArray
+@brief Make view on const @ref StaticArray
+
+Convenience alternative to calling @ref StaticArray::operator ArrayView<U>()
+explicitly. The following two lines are equivalent:
+
+@code{.cpp}
+const Containers::StaticArray<5, std::uint32_t> data;
+
+Containers::ArrayView<const std::uint32_t> a{data};
+auto b = Containers::arrayView(data);
+@endcode
+*/
+template<std::size_t size, class T> constexpr ArrayView<const T> arrayView(const StaticArray<size, T>& array) {
+    return ArrayView<const T>{array};
+}
+
+/** @relatesalso StaticArray
+@brief Make static view on @ref StaticArray
+
+Convenience alternative to calling @cpp StaticArray::operator StaticArrayView<size_, U>() @ce
+explicitly. The following two lines are equivalent:
+
+@code{.cpp}
+Containers::StaticArray<5, std::uint32_t> data;
+
+Containers::StaticArrayView<5, std::uint32_t> a{data};
+auto b = Containers::staticArrayView(data);
+@endcode
+
+@todoc Make it a real reference once Doxygen is sane
+*/
+template<std::size_t size, class T> constexpr StaticArrayView<size, T> staticArrayView(StaticArray<size, T>& array) {
+    return StaticArrayView<size, T>{array};
+}
+
+/** @relatesalso StaticArray
+@brief Make static view on const @ref StaticArray
+
+Convenience alternative to calling @cpp StaticArray::operator StaticArrayView<size_, U>() @ce
+explicitly. The following two lines are equivalent:
+
+@code{.cpp}
+const Containers::StaticArray<5, std::uint32_t> data;
+
+Containers::StaticArrayView<5, const std::uint32_t> a{data};
+auto b = Containers::staticArrayView(data);
+@endcode
+
+@todoc Make it a real reference once Doxygen is sane
+*/
+template<std::size_t size, class T> constexpr StaticArrayView<size, const T> staticArrayView(const StaticArray<size, T>& array) {
+    return StaticArrayView<size, const T>{array};
+}
+
+/** @relatesalso StaticArray
+@brief Reinterpret-cast a static array
+
+See @ref arrayCast(StaticArrayView<size, T>) for more information.
+*/
+template<class U, std::size_t size, class T> StaticArrayView<size*sizeof(T)/sizeof(U), U> arrayCast(StaticArray<size, T>& array) {
+    return arrayCast<U>(staticArrayView(array));
+}
+
+/** @overload */
+template<class U, std::size_t size, class T> StaticArrayView<size*sizeof(T)/sizeof(U), const U> arrayCast(const StaticArray<size, T>& array) {
+    return arrayCast<const U>(staticArrayView(array));
+}
+
+/** @relatesalso StaticArray
+@brief Static array size
+
+See @ref arraySize(ArrayView<T>) for more information.
+*/
+template<std::size_t size_, class T> constexpr std::size_t arraySize(const StaticArray<size_, T>&) {
+    return size_;
+}
+
+template<std::size_t size_, class T> template<class ...Args> StaticArray<size_, T>::StaticArray(DirectInitT, Args&&... args): StaticArray{NoInit} {
     for(T& i: _data) {
         /* MSVC 2015 needs the braces around */
         new(&i) T{std::forward<Args>(args)...};
     }
 }
 
-template<std::size_t size, class T> StaticArray<size, T>::~StaticArray() {
+template<std::size_t size_, class T> StaticArray<size_, T>::~StaticArray() {
     for(T& i: _data) i.~T();
 }
 
